%start Schedule

%avoid_insert "ID" "INT" "HEX_INT" "BIN_INT" "OCT_INT" "LABEL"
%expect-unused Unmatched 'UNMATCHED'

%left '||'
%left '&&'
%nonassoc '==' '!=' '<' '<=' '>' '>='
%left '+' '-'
%left '*' '/' '%'

%%

Schedule -> Vec<Partition> : PartitionList { $1 };

PartitionList -> Vec<Partition>
  : { vec![] }
  | PartitionList Partition { snoc($1, $2) }
  ;

Partition -> Partition
  : 'partition' Func '{' Labels '}' Directives
  { Partition { span : $span, func : $2, labels : $4, directs : $6 } };

Func -> Func
  : 'ID' { Func { span : $span, name : $span, args : None }}
  | 'ID' '::' '<' TypeExprs '>' { Func { span : $span, name : span_of_tok($1), args : Some($4) }}
  ;

Labels -> Vec<Span> : LabelsRev { rev($1) };
LabelsRev -> Vec<Span>
  : { vec![] }
  | 'LABEL' { vec![span_of_tok($1)] }
  | 'LABEL' ',' LabelsRev { cons(span_of_tok($1), $3) }
  ;

Directives -> Vec<Directive>
  : { vec![] }
  | Directives Directive { snoc($1, $2) }
  ;

Directive -> Directive
  : 'on' Device '{' Directives '}'
      { Directive::OnDevice { span : $span, device : $2, directs : $4 } }
  | 'if' Expr '{' Directives '}'
      { Directive::IfExpr { span : $span, cond : $2, directs : $4 } }
  | 'LABEL' '{' Commands '}'
      { Directive::Command { span : $span, label : span_of_tok($1), commands : $3 } }
  ;

Device -> Device
  : 'ID'
      { Device { span : $span, name : span_of_tok($1), args : vec![] } }
  | 'ID' '(' Exprs ')'
      { Device { span : $span, name : span_of_tok($1), args : $3 } }
  ;

Commands -> Vec<Command> : CommandsRev { rev($1) };
CommandsRev -> Vec<Command>
  : { vec![] }
  | Command { vec![$1] }
  | Command ',' CommandsRev { cons($1, $3) }
  ;

Command -> Command
  : 'ID'
      { Command { span : $span, name : span_of_tok($1), args : vec![] } }
  | 'ID' '(' Exprs ')'
      { Command { span : $span, name : span_of_tok($1), args : $3 } }
  ;

Exprs -> Vec<Expr> : ExprsRev { rev($1) };
ExprsRev -> Vec<Expr>
  : { vec![] }
  | Expr { vec![$1] }
  | Expr ',' ExprsRev { cons($1, $3) }
  ;

TypeExprs -> Vec<TypExp> : TypeExprsRev { rev($1) };
TypeExprsRev -> Vec<TypExp>
  : { vec![] }
  | TypeExp { vec![$1] }
  | TypeExp ',' TypeExprsRev { cons($1, $3) }
  ;

Expr -> Expr
  : 'ID'            { Expr::Id { span : $span }}
  | 'INT'           { Expr::Int { span : $span, base : IntBase::Dec }}
  | 'HEX_INT'       { Expr::Int { span : $span, base : IntBase::Hex }}
  | 'BIN_INT'       { Expr::Int { span : $span, base : IntBase::Bin }}
  | 'OCT_INT'       { Expr::Int { span : $span, base : IntBase::Oct }}
  | 'true'          { Expr::Bool { span : $span, value : true }}
  | 'false'         { Expr::Bool { span : $span, value : false }}
  | Expr '+' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Add, rhs : Box::new($3) }}
  | Expr '-' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Sub, rhs : Box::new($3) }}
  | Expr '*' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Mul, rhs : Box::new($3) }}
  | Expr '/' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Div, rhs : Box::new($3) }}
  | Expr '%' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Mod, rhs : Box::new($3) }}
  | Expr '==' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Eq, rhs : Box::new($3) }}
  | Expr '!=' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Ne, rhs : Box::new($3) }}
  | Expr '<' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Lt, rhs : Box::new($3) }}
  | Expr '<=' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Le, rhs : Box::new($3) }}
  | Expr '>' Expr   { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Gt, rhs : Box::new($3) }}
  | Expr '>=' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Ge, rhs : Box::new($3) }}
  | Expr '&&' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::And, rhs : Box::new($3) }}
  | Expr '||' Expr  { Expr::BinOp { span : $span, lhs : Box::new($1), op : BinOp::Or, rhs : Box::new($3) }}
  ;

TypeExp -> TypExp
  : '_'                   { TypExp::Wildcard { span : $span }}
  | 'ID'                  { TypExp::Id { span : $span }}
  | 'bool'                { TypExp::Primitive { span : $span, typ : Primitive::Bool }}
  | 'i8'                  { TypExp::Primitive { span : $span, typ : Primitive::I8 }}
  | 'i16'                 { TypExp::Primitive { span : $span, typ : Primitive::I16 }}
  | 'i32'                 { TypExp::Primitive { span : $span, typ : Primitive::I32 }}
  | 'i64'                 { TypExp::Primitive { span : $span, typ : Primitive::I64 }}
  | 'u8'                  { TypExp::Primitive { span : $span, typ : Primitive::U8 }}
  | 'u16'                 { TypExp::Primitive { span : $span, typ : Primitive::U16 }}
  | 'u32'                 { TypExp::Primitive { span : $span, typ : Primitive::U32 }}
  | 'u64'                 { TypExp::Primitive { span : $span, typ : Primitive::U64 }}
  | 'usize'               { TypExp::Primitive { span : $span, typ : Primitive::USize }}
  | 'f32'                 { TypExp::Primitive { span : $span, typ : Primitive::F32 }}
  | 'f64'                 { TypExp::Primitive { span : $span, typ : Primitive::F64 }}
  | 'void'                { TypExp::Primitive { span : $span, typ : Primitive::Void }}
  | '(' TypeExprs ')'     { TypExp::Tuple { span : $span, elems : $2 }}
  | '{' 'ID' '|' Expr '}' { TypExp::Set { span : $span, nm : span_of_tok($2), cond : $4 }}
  ;

Unmatched -> () : 'UNMATCHED' {};

%%

use cfgrammar::Span;
use lrlex::DefaultLexeme;

fn span_of_tok(t : Result<DefaultLexeme, DefaultLexeme>) -> Span {
  t.map_err(|_| ()).map(|l| l.span()).unwrap()
}

fn cons<A>(hd : A, mut tl : Vec<A>) -> Vec<A> {
  tl.push(hd);
  tl
}

fn snoc<A>(mut hd : Vec<A>, tl : A) -> Vec<A> {
  hd.push(tl);
  hd
}

fn rev<A>(mut lst : Vec<A>) -> Vec<A> {
  lst.reverse();
  lst
}

pub struct Partition {
  pub span    : Span,
  pub func    : Func,
  pub labels  : Vec<Span>,
  pub directs : Vec<Directive>,
}
pub struct Func { pub span : Span, pub name : Span, pub args : Option<Vec<TypExp>>, }
pub struct Device { pub span : Span, pub name : Span, pub args : Vec<Expr>, }
pub struct Command { pub span : Span, pub name : Span, pub args : Vec<Expr>, }

pub enum Expr {
  Id { span : Span },
  Int { span : Span, base : IntBase },
  Bool { span : Span, value : bool },
  BinOp { span : Span, lhs : Box<Expr>, op : BinOp, rhs : Box<Expr> },
}
pub enum TypExp {
  Wildcard { span : Span },
  Id { span : Span },
  Primitive { span : Span, typ : Primitive },
  Tuple { span : Span, elems : Vec<TypExp> },
  Set { span : Span, nm : Span, cond : Expr },
}
pub enum Directive {
  OnDevice { span : Span, device : Device, directs : Vec<Directive> },
  IfExpr   { span : Span, cond : Expr, directs : Vec<Directive> },
  Command  { span : Span, label : Span, commands : Vec<Command> },
}

pub enum IntBase { Bin, Oct, Dec, Hex }
pub enum BinOp { Add, Sub, Mul, Div, Mod, Eq, Ne, Lt, Le, Gt, Ge, And, Or }
pub enum Primitive { Bool, I8, I16, I32, I64, U8, U16, U32, U64, USize, F32, F64, Void }
